#include <LiquidCrystal.h>
#include "i2c.h"

#define LCD_SHIELD
// #define DEBUG

//Pin assignement
#ifdef LCD_SHIELD
  #define BTN_UP A1
  #define BTN_DOWN A2
  #define BTN_GO A3
  #define RS 8
  #define EN 9
  #define D4 4
  #define D5 5
  #define D6 6
  #define D7 7
#else
  #define BTN_UP 8
  #define BTN_DOWN 9
  #define BTN_GO 10
  #define RS 12
  #define EN 11
  #define D4 5
  #define D5 4
  #define D6 3
  #define D7 2
#endif
#define DEBOUNCE 50

//Functions prototypes
void loadCharBank(unsigned char bank);
bool checkAddress(unsigned char address);
void sniffAddresses();
void printSegment(unsigned char segment);
void printAll();

//Define the custom bitmaps for the LCD
const byte black0[8]={
  0x1F, 0x11, 0x15, 0x15, 0x15, 0x15, 0x11, 0x1F
};
const byte black1[8]={
  0x1F, 0x1B, 0x13, 0x1B, 0x1B, 0x1B, 0x11, 0x1F
};
const byte black2[8]={
  0x1F, 0x11, 0x1D, 0x11, 0x17, 0x17, 0x11, 0x1F
};
const byte black3[8]={
  0x1F, 0x11, 0x1D, 0x19, 0x1D, 0x1D, 0x11, 0x1F
};
const byte black4[8]={
  0x1F, 0x15, 0x15, 0x11, 0x1D, 0x1D, 0x1D, 0x1F
};
const byte black5[8]={
  0x1F, 0x11, 0x17, 0x11, 0x1D, 0x1D, 0x11, 0x1F
};
const byte black6[8]={
  0x1F, 0x11, 0x17, 0x17, 0x11, 0x15, 0x11, 0x1F
};
const byte black7[8]={
  0x1F, 0x11, 0x1D, 0x1D, 0x1B, 0x17, 0x17, 0x1F
};
const byte black8[8]={
  0x1F, 0x11, 0x15, 0x11, 0x11, 0x15, 0x11, 0x1F
};
const byte black9[8]={
  0x1F, 0x11, 0x15, 0x11, 0x1D, 0x1D, 0x11, 0x1F
};
const byte blackA[8]={
  0x1F, 0x11, 0x15, 0x11, 0x11, 0x15, 0x15, 0x1F
};
const byte blackB[8]={
  0x1F, 0x17, 0x17, 0x11, 0x15, 0x15, 0x11, 0x1F
};
const byte blackC[8]={
  0x1F, 0x11, 0x17, 0x17, 0x17, 0x17, 0x11, 0x1F
};
const byte blackD[8]={
  0x1F, 0x1D, 0x1D, 0x11, 0x15, 0x15, 0x11, 0x1F
};
const byte blackE[8]={
  0x1F, 0x11, 0x17, 0x11, 0x17, 0x17, 0x11, 0x1F
};
const byte blackF[8]={
  0x1F, 0x11, 0x17, 0x11, 0x17, 0x17, 0x17, 0x1F
};

//Constants for printing the addresses
const char header[17]="0123456789ABCDEF";

//Storage place for the scan results
unsigned short adresses[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
//Index for the addresses segment currently printed
unsigned char segment=0;
//Current bitmaps bank loaded for printing
unsigned char current_bank=0;

LiquidCrystal lcd(RS, EN, D4, D5, D6, D7);

void setup() {
  pinMode(BTN_UP, INPUT_PULLUP);
  pinMode(BTN_DOWN, INPUT_PULLUP);
  pinMode(BTN_GO, INPUT_PULLUP);
  i2cInit();
  lcd.begin(16, 2);
  lcd.clear();
  lcd.setCursor(2, 0);
  lcd.print("I2C Sniffer");
  lcd.setCursor(3, 1);
  lcd.print("Kosme 2018");
  delay(2000);
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Push Go to sniff");
  Serial.begin(115200);
}

void loop() {
  char btn = digitalRead(BTN_GO);
  if(!btn){
    delay(DEBOUNCE);
    if(digitalRead(BTN_GO==btn)){
      sniffAddresses();
      segment=0;
      printSegment(segment);
    }
  }
  btn = digitalRead(BTN_DOWN);
  if(!btn){
    delay(DEBOUNCE);
    if(digitalRead(BTN_DOWN==btn)){
      segment=(segment+1)&0x0F;//Limit the value to the range 0-15
      printSegment(segment);
    }
  }
  btn = digitalRead(BTN_UP);
  if(!btn){
    delay(DEBOUNCE);
    if(digitalRead(BTN_UP==btn)){
      segment=(segment-1)&0x0F;//Limit the value to the range 0-15
      printSegment(segment);
    }
  }
  //The serial printing can be ommitted for a portable device
  if(Serial.available()){
    unsigned char b=Serial.read();
    switch (b) {
      case 'k':
        sniffAddresses();
        printAll();
        break;
      default:
        break;
    }
  }
  delay(100);
}

//Check if an address is connected to a device
bool checkAddress(unsigned char address){
  bool res = FALSE;
  i2cSendStart(FALSE);
  i2cSendByte(address);
  i2cWaitForComplete();
  unsigned char status = i2cGetStatus();
  res=((status==0x18)||(status==0x40));
  if(address%2){ //Direccion de lectura
    i2cReceiveByte(FALSE);
    i2cWaitForComplete();
    i2cGetReceivedByte();
  }
  i2cSendStop();
  return(res);
}

//Scan all the addresses
void sniffAddresses(){
  for(unsigned char c=0;c<16;c++){
    unsigned short s=0;
    for(unsigned char b=0;b<16;b++){
      s=s>>1;
      unsigned char address = (c*0x10)+b;
      if(checkAddress(address)){
        s=s|0x8000;
      }
    }
    adresses[c]=s;
  }
}

//Serially print all the results of the scanning
void printAll(){
  Serial.print("\t ");
  for(unsigned char seg=0;seg<16;seg++){
    Serial.print(header[seg]);
    Serial.print(" ");
  }
  Serial.println();
  for(unsigned char s=0;s<16;s++){
    unsigned short mask=1;
    unsigned short seg = adresses[s];
    Serial.print("0x");
    Serial.print(s,HEX);
    Serial.print("0\t");
    for(unsigned char c=0;c<16;c++){
      if((seg&mask)==mask){
        if(c%2){//Read
          Serial.print(" R");
        }else{//Write
          Serial.print(" W");
        }
      }else{
        Serial.print(" -");
      }
      mask=mask<<1;
    }
    Serial.println();
  }
}

//Print on the LCD an address segment
void printSegment(unsigned char segment){
  if(segment>0xf){
    return;
  }
  unsigned short seg = adresses[segment];
  if(segment>7){//Address range between 0x80 and 0xF0
    loadCharBank(2);
  }else{//Address range between 0x00 and 0x70
    loadCharBank(1);
  }
  lcd.setCursor(0,0);
  lcd.clear();
  for(unsigned char c=0;c<0x10;c++){
    if(c==segment){
      if(segment<8){
        lcd.write(byte(c));
      } else{
        lcd.write(byte(c-8)); //lcd.write() parametter must be in range 0-7
      }
      #ifdef DEBUG
      Serial.print("H");
      #endif
    } else {
      lcd.print(header[c]);
      #ifdef DEBUG
      Serial.print(" ");
      #endif
    }
    Serial.print(header[c]);
  }
  Serial.println();
  lcd.setCursor(0,1);
  unsigned short mask=1;
  for(unsigned char c=0;c<0x10;c++){
    if((seg&mask)==mask){//Check if a device was detected on that address
      if(c%2){//It belongs to a Read address since they are always odd numbers
        lcd.print("R");
        #ifdef DEBUG
        Serial.print(" R");
        #endif
      }else{//It belongs to a Write address since they are always even numbers
        lcd.print("W");
        #ifdef DEBUG
        Serial.print(" W");
        #endif
      }
    }else{//No device was detected
      lcd.print("-");
      #ifdef DEBUG
      Serial.print(" -");
      #endif
    }
    mask=mask<<1;
  }
  #ifdef DEBUG
  Serial.println();
  #endif
}

//Put on the LCD memory the bold font characters.
//There are two banks because the LCD can only hold 8 custom characters at a time.
void loadCharBank(unsigned char bank){
  if((bank==current_bank)||(bank>2)){//If the needed bank is already loaded, don't load it again
    return;
  }
  if(bank==1){
    lcd.createChar(0, black0);
    lcd.createChar(1, black1);
    lcd.createChar(2, black2);
    lcd.createChar(3, black3);
    lcd.createChar(4, black4);
    lcd.createChar(5, black5);
    lcd.createChar(6, black6);
    lcd.createChar(7, black7);
  }else{
    lcd.createChar(0, black8);
    lcd.createChar(1, black9);
    lcd.createChar(2, blackA);
    lcd.createChar(3, blackB);
    lcd.createChar(4, blackC);
    lcd.createChar(5, blackD);
    lcd.createChar(6, blackE);
    lcd.createChar(7, blackF);
  }
  current_bank=bank;
}

